module dwin.sys.win32.com.InProcessServer;

/**
 * Contains code for creating In-Process COM Servers
 * 
 * 
 * Example Usage:
 * module inproctest;
 * 
 * version(build) {
 *     pragma(build_def, "EXPORTS");
 *     pragma(build_def, "DllRegisterServer");
 *     pragma(build_def, "DllUnregisterServer");
 *     pragma(build_def, "DllGetClassObject");
 *     pragma(build_def, "DllCanUnloadNow");
 * }
 * 
 * 
 * import dwin.sys.win32.com.InProcessServer;
 * import logger = ag.logger;
 * import tango.sys.win32.Types;
 * import tango.sys.win32.UserGdi;
 * import dwin.sys.win32.Types;
 * import dwin.sys.win32.Interfaces;
 * import dwin.sys.win32.com.Core;
 * import dwin.sys.win32.Registry;
 * import dwin.sys.win32.Wrapper;
 * 
 * class MyComObject : Implements!(IUnknown)
 * {
 * 	//	 {4F273070-A348-4ea2-84D7-E35313AEB8E4}
 * 	static Guid CLSID = { 0x4f273070, 0xa348, 0x4ea2, 0x84, 0xd7, 0xe3, 0x53, 0x13, 0xae, 0xb8, 0xe4 };
 * 	static char[] description = "A Test Class";
 * 	
 * 	static void register()
 * 	{
 * 		//No special registration required
 * 	}
 * 	
 * }
 * 
 * extern(Windows) int DllUnregisterServer()
 * {
 *	   return UnregisterServer!(MyComObject)(); 	
 * }
 * 
 * extern(Windows) int DllGetClassObject(ref Guid rclsid, ref Guid riid, void** ppv)
 * {
 * 	return GetClassObject!(MyComObject)(rclsid, riid, ppv); 	
 * }
 * 
 * extern(Windows) int DllRegisterServer()
 * {
 * 	return RegisterServer!(MyComObject)(); 	
 * }
 * 
 */

import tango.sys.win32.Types;
import tango.sys.win32.UserGdi;
import dwin.sys.win32.Types;
import dwin.sys.win32.Interfaces;
import dwin.sys.win32.com.Core;
import dwin.sys.win32.Registry;
import dwin.sys.win32.Wrapper;
import tango.stdc.stringz;

version(build){
    pragma(link, "kernel32");
}

extern (C) bool  rt_init( void delegate( Exception ) dg = null );
extern (C) bool  rt_term( void delegate( Exception ) dg = null ); 

private HANDLE moduleHANDLE_;

HANDLE getHInstance() {
  return moduleHANDLE_;
}

void setHInstance(HANDLE value) {
  moduleHANDLE_ = value;
}

char[] getLocation() {
	return getModuleFileName(moduleHANDLE_);
}

class ClassFactory(T) : Implements!(IClassFactory) 
{

    int CreateInstance(IUnknown pUnkOuter, ref Guid riid, void** ppvObject) 
    {
    	logger.info("CreateInstance");

        if (pUnkOuter !is null && riid != IUnknown.IID/*uuidof!(IUnknown)*/)
        {
        	return CLASS_E_NOAGGREGATION;
        }

        *ppvObject = null;
        int hr = E_OUTOFMEMORY;

        T obj = new T;
        obj.AddRef();
        if (obj !is null) 
        {
            hr = obj.QueryInterface(riid, ppvObject);
            obj.Release();
        }
    	logger.info("CreateInstance Complete");
        return hr;
    }

    int LockServer(int fLock) 
    {
    	logger.info("LockServer");
    	if (fLock)
    		lockModule();
    	else
    		unlockModule();
    	logger.info("LockServer Complete");
    	return S_OK;
    }
}

extern(Windows) int DllMain(HINSTANCE hInstance, uint dwReason, void* pvReserved) 
{
    if (dwReason == 1 /*DLL_PROCESS_ATTACH*/) {
      setHInstance(hInstance);

      rt_init();
      return 1;
    }
    else if (dwReason == 0 /*DLL_PROCESS_DETACH*/) 
    {
    	OutputDebugStringA("DLL_PROCESS_DETACH\n");
    	rt_term();
    	OutputDebugStringA("rt_term complete\n");
        return 1;
    }
    return 0;
}

extern(Windows)
int DllCanUnloadNow() {
	OutputDebugStringA("DllCanUnloadNow\n");
	
    if(getLockCount() == 0)
   	{
    	OutputDebugStringA("DLLCanUnloadNow getLockCount is 0!\n");
    	return S_OK; 
   	}
    return S_FALSE;
  }


bool registerCoClass(T)() {
    bool success;

    try {
      scope clsidKey = RegistryKey.classesRoot.createSubKey("CLSID\\" ~ T.CLSID.toString());
      if (clsidKey !is null) {
          clsidKey.setValue!(char[])("", T.description);
          //clsidKey.setValue!(char[])("", T.classinfo.name ~ " Class");

        scope subKey = clsidKey.createSubKey("InprocServer32");
        if (subKey !is null) {
          subKey.setValue!(char[])(null, getLocation());
          subKey.setValue!(char[])("ThreadingModel", "Apartment");

          scope progIDSubKey = clsidKey.createSubKey("ProgID");
          if (progIDSubKey !is null) {
            progIDSubKey.setValue!(char[])(null, T.classinfo.name);

            scope progIDKey = RegistryKey.classesRoot.createSubKey(T.classinfo.name);
            if (progIDKey !is null) {
              progIDKey.setValue!(char[])(null, T.classinfo.name ~ " Class");

              scope clsidSubKey = progIDKey.createSubKey("CLSID");
              if (clsidSubKey !is null)
                clsidSubKey.setValue!(char[])(null, T.CLSID.toString());
            }
          }
        }
      }

      success = true;
    }
    catch {
      success = false;
    }

    return success;
}

bool unregisterCoClass(T)() {
    bool success;

    try {
      scope clsidKey = RegistryKey.classesRoot.openSubKey("CLSID");
      if (clsidKey !is null)
        clsidKey.deleteSubKeyTree(T.CLSID.toString());

      RegistryKey.classesRoot.deleteSubKeyTree(T.classinfo.name);

      success = true;
    }
    catch {
      success = false;
    }

    return success;
}

int GetClassObject(T ...)(ref Guid rclsid, ref Guid riid, void** ppv) 
	{
		logger.info("GetClassObject");
	    int hr = CLASS_E_CLASSNOTAVAILABLE;
	    *ppv = null;
	
	    foreach(coclass; T)
	    {
			if (rclsid == coclass.CLSID/*uuidof!(coclass)*/) 
			{
		 	    IClassFactory factory = new ClassFactory!(coclass);
			    if (factory is null) return E_OUTOFMEMORY;
			    scope(exit) factory.Release();//tryRelease(factory);
			
			    hr = factory.QueryInterface(riid, ppv);
			}
	    }
		logger.info("GetClassObject Complete");
	    return hr;
	}
	
int RegisterServer(T ...)() 
	{
	    bool success;
	
	    foreach(coclass; T)
	    {
	        success = registerCoClass!(coclass)();
	        coclass.register();
	    }
	
	    return success ? S_OK : SELFREG_E_CLASS;
	}
	
int UnregisterServer(T ...)() 
	{
	    bool success;
	
	    foreach(coclass; T)
	    {
	        success = unregisterCoClass!(coclass)();
	        coclass.unregister();
	    }
	
	    return success ? S_OK : SELFREG_E_CLASS;
	}
